#ifndef RNASTRUCTURE_H_
#define RNASTRUCTURE_H_


#include "BioMolecule.hh"
#include "AllowedBasePairs.hh"

namespace ell
{
		/**
		 * An object of the class RNAStructure represents a RNA as a
		 * BioMolecule.
		 * 
		 * An object of this class is allowed to have only nested bonds.
		 * 
		 * @author Martin Mann, Sebastian Will, Andreas Richter
		 */
	class RNAStructure : public BioMolecule
	{
	protected:
			//! Minimal length of an RNA loop
		static const unsigned int MIN_LOOP_LENGTH = 3;

			//! RNA sequence
		Sequence*		rnaSeq;
		
			//! RNA structure in bracket notation
		Structure*		rnaStructBracketNot;
		
			/*! Mapping from index of an opening bond to the
			 *  index of the corresponding closing bond
			 *  Entry at index i is index of the closing bond
			 *  and zero if index i is not an opening bond
			 */
		std::vector<unsigned int> rnaBonds;
		
			/*! A pointer to an BasePair object containing the
			 * valid RNA base pairs
			 */
		const AllowedBasePairs* 	bPair;

	public:
			/*! @param rnaSeqStr is the RNA sequence
			 * @param _rnaStructBracketNot	is the RNA structure in bracket
			 *  							notation.
			 * @param _bPair is the handler for the allowed base pairs.
			 * 
			 *  It is explicitly allowed to construct invalid
			 *  RNAStructures, but each bond has to be
			 *  complete (each opening has a closing bracket).
			 */
		RNAStructure(	const std::string& rnaSeqStr, 
						const Structure& _rnaStructBracketNot, 
						const AllowedBasePairs* const _bPair);
			     
		RNAStructure(const RNAStructure& rnaStruct);
		
		virtual ~RNAStructure();
		
		RNAStructure&	operator= (const RNAStructure& rnaStruct2);
		bool	 		operator== (const RNAStructure& rnaStruct2) const;
		bool	 		operator!= (const RNAStructure& rnaStruct2) const;

			/*! Returns whether the RNA structure is in valid
			 *  bracket notation (each opening has a closing bracket).
			 */
		bool			hasValidStructure() const;
		
			/*! Returns whether all RNA base pairs of an
			 *  RNAStructure object are valid.
			 */
		bool			hasValidBasePairs() const;
		
			/*! Returns whether all loops of an RNAStructure object
			 *  have at least a length of MIN_LOOP_LENGTH.
			 */
		bool			hasMinLoopLength() const;
		
			/*! Returns whether a base pair between positions
			 *  first and second is valid.
			 */
		bool isAllowedBasePair(unsigned int first, unsigned int second) const;

			//! Constant that represents a invalid structure index position.
		static const unsigned int INVALID_INDEX;
		
			//! Returns  the corresponding closing bond index to the given
			//! opening index
			//! 
			//! Returns INVALID_INDEX in error case, if there doesnt exist 
			//! a closing bond.
		unsigned int getClosingBond(unsigned int openingBond) const { 
			return rnaBonds[openingBond]; 
		}
		
		unsigned int getMinLoopLength() const {
			return MIN_LOOP_LENGTH; 
		}

	// abstract functions (BioMolecule)
	
		virtual Sequence	getSequence() const;
		
			//! Returns the RNA structure in bracket notation.
		virtual Structure	getStructure() const;
		
			/*! Returns true if an RNAStructure object has a
			 *  valid structure with valid base pairs and
			 *  loops of a length of at least MIN_LOOP_LENGTH.
			 */
		virtual bool 		isValid() const;
		
			/*! Returns a combination of bracket notation 
			 *  and sequence information of this RNA.
			 * 
			 *  BRACKETNOTATION(SEQUENCE) */
		virtual std::string	getStringRepresentation() const;
	};
	
} // namespace ell

#endif /*RNASTRUCTURE_H_*/
