#include <ell/assertell.hh>
#include <ell/RNAFreeEnergySingleMoveState.hh>
#include <ell/assertell.hh>
#include <cmath>


namespace ell
{
	RNAFreeEnergySingleMoveState::RNAFreeEnergySingleMoveState(
	 			const std::string& rnaSeqStr, 
	 			const Structure& _rnaStructBracketNot,
	 			const AllowedBasePairs* const _bPair)
	 :	rnaFreeEnergy(RNAFreeEnergy(rnaSeqStr, _rnaStructBracketNot, _bPair)) 
	{
		assertell(rnaFreeEnergy.isValid(),
		 "RNAFreeEnergySingleMoveState has to be a valid state.");
	}

	RNAFreeEnergySingleMoveState::RNAFreeEnergySingleMoveState(
				const RNAFreeEnergySingleMoveState& rnaFESMS)
	 : rnaFreeEnergy(rnaFESMS.rnaFreeEnergy) {
	}

	RNAFreeEnergySingleMoveState::~RNAFreeEnergySingleMoveState() {
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState::firstSingleMove() const {
		return searchForNextSingleMove(0,1);
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: nextSingleMove( 
			const SingleMove lastSingleMove) const 
	{
		// if lastSingleMove.first is an opening bond, this bond was already
		// modified in the last move
		// closing bonds at lastSingleMove.first are not for interest
		if (rnaFreeEnergy.getStructure().at(lastSingleMove.first) != '.' ) {
			return searchForNextSingleMove(lastSingleMove.first+1,
				lastSingleMove.first+2);
		}
		else {
			return searchForNextSingleMove(lastSingleMove.first,
				lastSingleMove.second+1);
		}
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: searchForNextSingleMove(	unsigned int pos1, 
								unsigned int pos2) const 
	{
		assertell(pos1 < pos2,
		 "RNA single move search pos 1 has to be less than pos 2");
		
		// distance between pos1 and pos2 has to be at least
		//  MIN_LOOP_LENGTH
		while ( (pos1+rnaFreeEnergy.getMinLoopLength())
				 < rnaFreeEnergy.getStructure().size()) 
		{
			switch (rnaFreeEnergy.getStructure().at(pos1)) {
				case ')' : break;
				case '(' : pos2 = rnaFreeEnergy.getClosingBond(pos1);
					   // existing bond found -> exit loop
					   return SingleMove(pos1,pos2);
					   break;
				case '.' : {
					for(pos2=rnaFreeEnergy.nextCompatibleSingleMovePos(pos2-1);
					    pos2 < rnaFreeEnergy.getStructure().size();
					    pos2 = rnaFreeEnergy.nextCompatibleSingleMovePos(pos2))
					{
						if ((pos2 - pos1) > rnaFreeEnergy.getMinLoopLength() &&
						    rnaFreeEnergy.isAllowedBasePair(pos1, pos2)) {
							// found bases which are able to pair
							// -> exit loop
							return SingleMove(pos1,pos2);
							break;
						}
					   }
					   break;
					}
			}
			// loop wasn't exited since no next valid single move was found
			pos1++;
			pos2 = pos1;
		}
		// no next valid single move found
		return SingleMove(0,RNAStructure::INVALID_INDEX);
	}

	bool	
	RNAFreeEnergySingleMoveState::operator== (const State& state2) const {
		return rnaFreeEnergy == (dynamic_cast<
			const RNAFreeEnergySingleMoveState&>(state2)).rnaFreeEnergy;
	}

	bool
	RNAFreeEnergySingleMoveState::operator!= (const State& state2) const {
		return rnaFreeEnergy != (dynamic_cast<
			const RNAFreeEnergySingleMoveState&>(state2)).rnaFreeEnergy;
	}

	bool
	RNAFreeEnergySingleMoveState::operator< (const State& state2) const {
		RNAFreeEnergySingleMoveState rnaState2 =
			dynamic_cast<const RNAFreeEnergySingleMoveState&>(state2);
		return rnaFreeEnergy.getStructure() <
				rnaState2.rnaFreeEnergy.getStructure() ||
				(rnaFreeEnergy.getStructure() ==
					rnaState2.rnaFreeEnergy.getStructure() &&
				rnaFreeEnergy.getSequence() <
					rnaState2.rnaFreeEnergy.getSequence());
	}

	int	
	RNAFreeEnergySingleMoveState::getFitness() const {
			// rounds fitness and multiplies it with 100
		double fitness = 100 * rnaFreeEnergy.getEnergy();
		if (fitness >= 0)
			return static_cast<int>(floor(fitness + 0.5));
		else
			return static_cast<int>(ceil(fitness - 0.5));
	}

	int	
	RNAFreeEnergySingleMoveState::getMinimalDistance(
		 const State& _state2) const {
		RNAFreeEnergySingleMoveState state2 = 
		 dynamic_cast<const RNAFreeEnergySingleMoveState&>(_state2);

		assertell(state2.rnaFreeEnergy.getStructure().size() ==
		 rnaFreeEnergy.getStructure().size(),
		"Length of both RNA has to be equal to compute the distance.");
			
		int distance = 0;
		for (unsigned int i = 0; i < rnaFreeEnergy.getStructure().size(); i++){
			// bond maintained or bond opened and replaced by another one
			if (rnaFreeEnergy.getStructure().at(i) == '(' &&
			     state2.rnaFreeEnergy.getStructure().at(i) == '(') {
				// bond opened and replaced by another one
				if (rnaFreeEnergy.getClosingBond(i) !=
				     state2.rnaFreeEnergy.getClosingBond(i)) {
					distance += 2;
				}
			}
			// bond opened or closed
			else if (rnaFreeEnergy.getStructure().at(i) == '(' || 
				  state2.rnaFreeEnergy.getStructure().at(i) == '(') {
				distance += 1;
			}
		}
		return distance;
	}

	RNAFreeEnergySingleMoveState* 
	RNAFreeEnergySingleMoveState::clone() const {
		return new RNAFreeEnergySingleMoveState(*this);
	}
	
	std::string 
	RNAFreeEnergySingleMoveState::getStateStringRepresentation()
		     const{
		return rnaFreeEnergy.getStringRepresentation();
	}

	State::NeighborListPtr 
	RNAFreeEnergySingleMoveState::getNeighborList() const {
		return NeighborListPtr(new NeighborList(this));
	}

	State::NeighborListPtr
	RNAFreeEnergySingleMoveState::getRandomNeighborList()
				const {
		return NeighborListPtr(new RandomNeighborList(this));
	}

	// NeighborList
	RNAFreeEnergySingleMoveState::NeighborList::NeighborList(
	 			const RNAFreeEnergySingleMoveState* _origin) 
	 : origin(_origin) 
	{
	}

	RNAFreeEnergySingleMoveState::NeighborList::~NeighborList() {
	}

	State* 
	RNAFreeEnergySingleMoveState::NeighborList::first(
		      State::NeighborList::ItState** itstate) const {
		if (*itstate != NULL) delete *itstate;

		ItState* myitstate = new ItState();
		*itstate = myitstate;
		
		RNAFreeEnergySingleMoveState* rnaState = NULL;

		SingleMove firstMove = origin->firstSingleMove();
		if (firstMove != SingleMove(0,RNAStructure::INVALID_INDEX)) {
			rnaState = new RNAFreeEnergySingleMoveState(*origin);
			rnaState->rnaFreeEnergy.applySingleMove(firstMove);
			myitstate->lastMove = firstMove;
		}
		return rnaState;
	}

	State* 
	RNAFreeEnergySingleMoveState::NeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const {
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);

		// find next move
		SingleMove nextMove = origin->nextSingleMove(myitstate->lastMove);
		// no more moves possible
		if (nextMove == SingleMove(0,RNAStructure::INVALID_INDEX)) {
			return NULL;
		}
		else {
			RNAFreeEnergySingleMoveState* rnaState =
			 dynamic_cast<RNAFreeEnergySingleMoveState*>(elem);
			// undo last move
			rnaState->rnaFreeEnergy.applySingleMove(myitstate->lastMove);
			// perform next move
			rnaState->rnaFreeEnergy.applySingleMove(nextMove);
			myitstate->lastMove = nextMove;
			return rnaState;
		}
	}

	// RandomNeighborList
	RNAFreeEnergySingleMoveState::RandomNeighborList::RandomNeighborList(
	 		const RNAFreeEnergySingleMoveState* _origin)
	 : origin(_origin) 
	{
	}

	RNAFreeEnergySingleMoveState::RandomNeighborList::~RandomNeighborList() {
	}

	unsigned int 
	RNAFreeEnergySingleMoveState::RandomNeighborList::
		      getRandomNum(unsigned int max) const {
		return (random() % max);
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: RandomNeighborList::getOrderedSingleMove(
			unsigned int pos1,
	 		unsigned int pos2) const 
	{
		return (pos1 < pos2) ?
			SingleMove(pos1, pos2) : SingleMove(pos2, pos1);
	}

	void 
	RNAFreeEnergySingleMoveState::RandomNeighborList::switchMode(
			ItState* itstate) const 
	{
		ItState* myitstate = static_cast<ItState*>(itstate);

		SingleMove nextMove = origin->firstSingleMove();

		// loop while valid moves exist
		while (nextMove != SingleMove(0,RNAStructure::INVALID_INDEX)) {
			// add all valid and unchosen moves to unchosenValidMoves
			if (myitstate->chosenMoves.find(nextMove) == 
			     myitstate->chosenMoves.end()) {
				myitstate->unchosenValidMoves.push_back(nextMove);
			}
			nextMove = origin->nextSingleMove(nextMove);
		}
	}

	State* 
	RNAFreeEnergySingleMoveState::RandomNeighborList::first(
		      State::NeighborList::ItState** itstate) const 
	{
		if (*itstate != NULL) delete *itstate;

		*itstate = new ItState();
		RNAFreeEnergySingleMoveState* rnaState = 
		 new RNAFreeEnergySingleMoveState(*origin);

		return next(*itstate, rnaState);
	}

	State* 
	RNAFreeEnergySingleMoveState::RandomNeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const 
	{
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);
		RNAFreeEnergySingleMoveState* rnaState =
		 dynamic_cast<RNAFreeEnergySingleMoveState*>(elem);

		if (myitstate->switchModeValue > 0) {
			// positions the move will be applied to
			unsigned int pos1, pos2;
			SingleMove nextMove;
			bool foundValidMove = false;

			while (!foundValidMove) {
				pos1 = getRandomNum(
					origin->rnaFreeEnergy.getStructure().size());
				pos2 = getRandomNum(
					origin->rnaFreeEnergy.getStructure().size());
				if (pos1 == pos2) {
					continue;
				}
				nextMove = getOrderedSingleMove(pos1, pos2);
				// move has already been chosen
				if (!myitstate->chosenMoves.insert(nextMove).second) {
					myitstate->switchModeValue--;
					if (myitstate->switchModeValue == 0) {
						switchMode(myitstate);
						break;
					}
				}
				// move hasn't been chosen yet and is valid
				else if (origin->rnaFreeEnergy.
					 isValidSingleMove(nextMove)) {
					*rnaState = *origin;
					rnaState->rnaFreeEnergy.applySingleMove(
					 nextMove);
					foundValidMove = true;
				}	
			}
		}
		if (myitstate->switchModeValue == 0) {
			unsigned int n = myitstate->unchosenValidMoves.size();
			if (n == 0) {
				return NULL;
			}
			else {
				unsigned int i = getRandomNum(n);
				*rnaState = *origin;
				rnaState->rnaFreeEnergy.applySingleMove(
				 myitstate->unchosenValidMoves[i]);
				myitstate->unchosenValidMoves[i] =
				 myitstate->unchosenValidMoves[n-1];
				myitstate->unchosenValidMoves.resize(n-1);
			}
		}
		return rnaState;
	}

} // namespace ell
