#include <ell/assertell.hh>
#include <ell/RNAStructure.hh>
#include <ell/assertell.hh>
#include <stack>

namespace ell
{

	const unsigned int RNAStructure::INVALID_INDEX = INT_MAX;
//////////////////////////////////////////////////////////
// RNAStructure
//////////////////////////////////////////////////////////

	RNAStructure::RNAStructure(	const std::string& rnaSeqStr,
	 							const Structure& _rnaStructBracketNot,
	 							const AllowedBasePairs* const _bPair)
	 :	rnaSeq(NULL), 
	 	rnaStructBracketNot(new Structure(_rnaStructBracketNot)),
		bPair(_bPair) 
	{
		assertell(bPair != NULL, "bPair is not allowed to be NULL.");
		assertell(bPair->getAlphabet()->isAlphabetString(rnaSeqStr),
		 "RNA sequence contains characters which are not in the alphabet.");

		// std::stack openingBonds contains opening bonds to create
		// vector rnaBonds during intitialisation
		std::stack<unsigned int>	openingBonds;

		rnaSeq = new Sequence(bPair->getAlphabet()->getSequence(rnaSeqStr));

		assertell(rnaSeq->size() == rnaStructBracketNot->size(),
		 "RNA sequence and structure have to have the same length.");
	
		rnaBonds = std::vector<unsigned int>
			   (rnaStructBracketNot->size(), RNAStructure::INVALID_INDEX);
		for (size_t i=0; i <  rnaStructBracketNot->size(); i++) {
			if ((*rnaStructBracketNot)[i] == '(') {
				openingBonds.push(i);
			}
			else if ((*rnaStructBracketNot)[i] == ')') {
				assertell(!openingBonds.empty(),
				 "Closing without opening bond.");
				// check whether pairing of bases allowed	
				rnaBonds[openingBonds.top()] = i;
				openingBonds.pop();
			}
		}
		assertell(openingBonds.empty(), "Opening without closing bond.");
	}

	RNAStructure::RNAStructure(const RNAStructure& rnaStruct)
	 : rnaSeq(new Sequence(*rnaStruct.rnaSeq)),
	   rnaStructBracketNot(new Structure(*rnaStruct.rnaStructBracketNot)),
	   rnaBonds(rnaStruct.rnaBonds), bPair(rnaStruct.bPair) 
	{
		assertell(bPair != NULL, "bPair is not allowed to be NULL.");
		assertell(rnaSeq->size() == rnaStructBracketNot->size(),
		 "RNA sequence and structure have to have the same length.");
	}

	RNAStructure::~RNAStructure() {
		delete rnaSeq; rnaSeq = NULL;
		delete rnaStructBracketNot; rnaStructBracketNot = NULL;
	}

	bool 		
	RNAStructure::isAllowedBasePair(	unsigned int first,
			 							unsigned int second)  const 
	{
		if (first == RNAStructure::INVALID_INDEX || second == RNAStructure::INVALID_INDEX)
			return false;
		assertell(first < rnaSeq->size() && second < rnaSeq->size(),
		 "First or second is not a base position in the RNA sequence.");
		return bPair->allowedBasePair((*rnaSeq)[first], (*rnaSeq)[second]);
	}

	RNAStructure& 	
	RNAStructure::operator= (const RNAStructure& rnaStruct2) {
		if (this != &rnaStruct2) {
			delete rnaSeq;
			rnaSeq = new Alphabet::Sequence(*rnaStruct2.rnaSeq);
			delete rnaStructBracketNot;
			rnaStructBracketNot = new
				Structure(*rnaStruct2.rnaStructBracketNot);
			bPair = rnaStruct2.bPair;
			rnaBonds = rnaStruct2.rnaBonds;	
		}
		return *this;
	}

	bool	 	
	RNAStructure::operator== (const RNAStructure& rnaStruct2)
			 const {
		return *rnaSeq == *(rnaStruct2.rnaSeq)
			&& *rnaStructBracketNot == *(rnaStruct2.rnaStructBracketNot)
			&& rnaBonds == rnaStruct2.rnaBonds
			&& *bPair == *(rnaStruct2.bPair);
	}

	bool	 	
	RNAStructure::operator!= (const RNAStructure& rnaStruct2)
			 const {
		return *rnaSeq != *(rnaStruct2.rnaSeq)
			|| *rnaStructBracketNot != *(rnaStruct2.rnaStructBracketNot)
			|| rnaBonds != rnaStruct2.rnaBonds
			|| *bPair != *(rnaStruct2.bPair);
	}

	bool		
	RNAStructure::hasValidStructure() const{
		int count = 0;	// counter for open bonds
		// each opening bracket has to have a closing bracket
		for (size_t i=0; i < rnaStructBracketNot->size(); i++) {
			if ((*rnaStructBracketNot)[i] == '(') {
				count++;
			}
			else if ((*rnaStructBracketNot)[i] == ')') {
				count--;
			}
			if (count < 0) {
				return false;
			}
		}
		return (count == 0);
	}

	bool		
	RNAStructure::hasValidBasePairs() const{
		for (size_t i=0; i < rnaBonds.size(); i++) {
			// existing bond has to be an allowed base pair
			if (rnaBonds[i] != INVALID_INDEX
			    && !isAllowedBasePair(i, rnaBonds[i])) {
				return false;
			}
		}
		return true;
	}	

	bool		
	RNAStructure::hasMinLoopLength() const {
		for (size_t i=0; i < rnaBonds.size(); i++) {
			// existing bond has to have at least MIN_LENGTH_LOOP
			if (rnaBonds[i] != INVALID_INDEX
			    && (rnaBonds[i]-i) <= MIN_LOOP_LENGTH) {
				return false;
			}
		}
		return true;
	}

	Sequence	
	RNAStructure::getSequence() const {
		return *rnaSeq;
	}

	Structure	
	RNAStructure::getStructure() const {
		return *rnaStructBracketNot;
	}

	bool		
	RNAStructure::isValid() const{
		return (rnaSeq->size() == rnaStructBracketNot->size())
			&& hasValidStructure() && hasMinLoopLength()
			&& hasValidBasePairs();
	}

	std::string	
	RNAStructure::getStringRepresentation() const {
		return (*rnaStructBracketNot+"("+
			(bPair->getAlphabet()->getString(*rnaSeq))+")");
	}

} // namespace ell
